#include "learning.h"


Clearning::Clearning(void)
{
}

Clearning::~Clearning(void)
{
}

int Clearning::Init(){

	int i;

	m_feature = new double *[FILE_NUM];
	for(i=0; i<FILE_NUM; i++){

		printf("Preparing Memory : %7d / %d\r", i, FILE_NUM);
		m_feature[i] = new double[ALL_FEATURE];
	}
	printf("\nPrepare memory termination\n");

	return 0;
}

//Calculated probability density distribution
int Clearning::Comp_Hist(int number,double lut[][2]){
	int i;
	int tmp;
	double sump,sumn;
	sump = sumn = 0.0;

	//Initialization
	for(i=0;i<BIN;i++){
		lut[i][POS] = 0.0;
		lut[i][NEG] = 0.0;
	}

	for(i=0;i<FILE_NUM;i++){
		//Calculate the number of bins
		tmp = (int)(m_feature[i][number] * BIN);
		if(tmp >= BIN){
			tmp = BIN - 1;
		}
		//Added to the histogram
		if(i<POS_FILE){
			lut[tmp][POS] += m_weight[i];
		}else{
			lut[tmp][NEG] += m_weight[i];
		}
	}

	//Smoothing of probability density
	SmoothingLUT(lut);

	//Calculate the sum of the probability density distribution of each class to perform normalization
	for(i=0;i<BIN;i++){
		sump += lut[i][POS];
		sumn += lut[i][NEG];
	}

	//Normalization
	for(i=0;i<BIN;i++){
		lut[i][POS] = lut[i][POS] / sump;
		lut[i][NEG] = lut[i][NEG] / sumn;
	}

	return 0;
}

//Smoothing of probability density
int Clearning::SmoothingLUT(double lut[][2]){
	int i;
	double lut2[BIN][2];

	//Copies of the probability density distribution
	for(i = 0;i < BIN; i++){
		lut2[i][POS] = lut[i][POS];
		lut2[i][NEG] = lut[i][NEG];
	}

	//Smoothing of probability density
	for(i = 0; i < BIN; i++){
		if(i == 0){
			lut[i][POS] = (lut2[i][POS] + lut2[i+1][POS]) / 2.0;
			lut[i][NEG] = (lut2[i][NEG] + lut2[i+1][NEG]) / 2.0;
		}else if(i == 1){
			lut[i][POS] = (lut2[i-1][POS] + lut2[i][POS] + lut2[i+1][POS] + lut2[i+2][POS]) / 4.0;
			lut[i][NEG] = (lut2[i-1][NEG] + lut2[i][NEG] + lut2[i+1][NEG] + lut2[i+2][NEG]) / 4.0;
		}else if(i == BIN - 1){
			lut[i][POS] = (lut2[i-1][POS] + lut2[i][POS]) / 2.0;
			lut[i][NEG] = (lut2[i-1][NEG] + lut2[i][NEG]) / 2.0;
		}else if(i == BIN - 2){
			lut[i][POS] = (lut2[i-2][POS] + lut2[i-1][POS] + lut2[i][POS] + lut2[i+1][POS]) / 4.0;
			lut[i][NEG] = (lut2[i-2][NEG] + lut2[i-1][NEG] + lut2[i][NEG] + lut2[i+1][NEG]) / 4.0;
		}else{
			lut[i][POS] = (lut2[i-2][POS] + lut2[i-1][POS] + lut2[i][POS] + lut2[i+1][POS] + lut2[i+2][POS]) / 5.0;
			lut[i][NEG] = (lut2[i-2][NEG] + lut2[i-1][NEG] + lut2[i][NEG] + lut2[i+1][NEG] + lut2[i+2][NEG]) / 5.0;
		}
	}

	return 0;
}

//The output of weak classifier
double Clearning::Prediction(int number,int count,double lut[][2]){

	double h;
	int tmp;

	//Calculation of the bin
	tmp = (int)(m_feature[count][number] * BIN);
	if(tmp >= BIN){
		tmp = BIN - 1;
	}

	//Reliability calculation
	h = 0.5 * log((lut[tmp][POS] + EPSILON)/(lut[tmp][NEG] + EPSILON));

	return h;
}

//Calculate an estimate of the weak classifier
double Clearning::Classification(double lut[][2]){

	int i;
	double w;
	double z;

	//Initialization error rate
	z = 0.0;
	for(i = 0; i < BIN; i++){
		w = lut[i][POS] * lut[i][NEG];
		z += sqrt(w);
	}
	Evaluated_Value = z;
	z = 2.0 * z;
	return z;
}

//Weak classifiers selected by the evaluation
int Clearning::SelectClassifier(Classifier *cl,double lut[][2]){
	
	double minerror = 1000.0;
	double z;
	int number;
	int i;

	cl->number = 0;

	for(i = 0; i < ALL_FEATURE; i++){
		Comp_Hist(i,lut);
		z = Classification(lut);
		if(minerror > z){
			number = i;
			minerror = z;
		}	
	}
	cl->number = number;
	cl->z;

	//Identification of the selected weak classifiers
	Comp_Hist(cl->number,lut);
	cl->z = Classification(lut);

	return 0;
}

//Update weights for training data
int Clearning::DataWeightUpdate(int number,double lut[][2]){

	double all_weight = 0.0;		//Sum of all weights
	double c;
	int i;

	for(i = 0; i < FILE_NUM; i++){
		//Calculate the probability density distribution reliability
		c = Prediction(number,i,lut);
		if(i < POS_FILE){
			m_weight[i] = m_weight[i] * exp(-c);
		}else{
			m_weight[i] = m_weight[i] * exp(c);
		}	
		all_weight += m_weight[i];
		
	}
	//Normalized so that the sum of the weights is 1
	for(i = 0; i < FILE_NUM; i++){
		m_weight[i] = m_weight[i] / all_weight;
	}
	return 0;
}

//Writing the final hypothesis
int Clearning::WrightData(Classifier *cl,double lut[][2],int round){

	FILE	*fp;
	char fname[200];

	sprintf_s(fname, "./OUTPUT/boost.dat");
	if(( fp = fopen(fname,"a")) == NULL){
		printf("Can not LUT open error.\n");
		exit(1);
	}

	fprintf_s(fp, "%4d\n",cl->number);
	fclose(fp);

	sprintf_s(fname, "./OUTPUT/EvaluatedValue.dat");
	if(( fp = fopen(fname,"a")) == NULL){
		printf("Can not LUT open error.\n");
		exit(1);
	}

	fprintf_s(fp, "%lf\n", Evaluated_Value);
	fclose(fp);

#ifdef TEST 

	int i;

	m_number[round] = cl->number;
	for(i=0;i<BIN;i++){
		m_lut[round][i][POS] = lut[i][POS];
		m_lut[round][i][NEG] = lut[i][NEG];
	}

#endif 
	return 0;
}

//Calculate the initial value of weights for the training sample
int Clearning::DataWeightInit(){
	int i;
	for(i = 0; i < FILE_NUM; i++){
		if(i < POS_FILE){
			m_weight[i] = 1.0/((double)POS_FILE * 2.0);
		}else{
			m_weight[i] = 1.0/((double)NEG_FILE * 2.0);
		}
	}
	return 0;
}

//Features stock
int Clearning::FeatureStock(double hog_descriptor[ALL_FEATURE],int count){
	int i;
	for(i = 0; i < ALL_FEATURE; i++){
		m_feature[count][i] = hog_descriptor[i];
	}
	return 0;
}

int Clearning::OutPutHist(double lut[][2],int leran){
	int j;
	char fname[200];
	FILE* f;


	//Histogram Export to Files
	sprintf(fname,"%sf_%07d.dat",OUTPUT_DIR,leran);
	if((f = fopen(fname,"w")) == NULL){
		printf("Can Not File Open\n");
		exit(1);
	}
		
	for(j=0;j<BIN;j++){
		fprintf(f,"%d %f %f\n",j,lut[j][POS],lut[j][NEG]);
	}

	fclose(f);

	return 0;
}

//#ifdef TEST 

double Clearning::Prediction2(int number,int count){

	double c;
	int tmp;

	//Calculation of the bin
	tmp = (int)(m_feature[count][m_number[number]] * BIN);
	if(tmp >= BIN){
		tmp = BIN - 1;
	}


	//Reliability calculation
	c = 0.5 * log((m_lut[number][tmp][POS] + EPSILON)/(m_lut[number][tmp][NEG] + EPSILON));

	return c;
}

//Classification of training samples
int Clearning::ClassifierTest(int learn){
	int i,j;
	double p,q;
	int fcount = 0;
	int pcount = 0;
	int ffcount=0;
	double pr,fpr;
	p=0,q=0;

	for(i=0;i<FILE_NUM;i++){
		p=0.0;
		q=0.0;
		for(j=0;j<=learn;j++){
			p += Prediction2(j,i);
		}

		if(p >= TH){
			if(i < POS_FILE){
				pcount++;
			}else{
				fcount++;
			}
		}
	}
	pr = (double)pcount/POS_FILE;
	fpr = (double)fcount/(FILE_NUM - POS_FILE);

	printf("Ture positive : pr%f%\nFalse positive fpr%f%\n",pr,fpr);	

	return 0;
}

//#endif

//AdaBoost runing
int Clearning::RealAdaBoost(){

	int round;
	double lut[BIN][2];
	Classifier	cl;			//Weak hypothesis

	//Initialization of weights for the data
	DataWeightInit();

	//Learning
	for(round = 0; round < LEARNING_NUM; round++){

		printf("Round : %d\n",round + 1);

		//Weak hypothesis selection
		SelectClassifier(&cl,lut);

		//Histogram Export
		OutPutHist(lut,round);

		//Update weights for training data
		DataWeightUpdate(cl.number,lut);

		//Writing the final hypothesis
		WrightData(&cl,lut,round);
//#ifdef TEST 
		//Classification of training samples
		ClassifierTest(round);	
//#endif
	}

	return 0;
}
